/*****************************************************************************\
|                                               ( )_  _                       |
|    _ _    _ __   _ _    __    ___ ___     _ _ | ,_)(_)  ___   ___     _     |
|   ( '_`\ ( '__)/'_` ) /'_ `\/' _ ` _ `\ /'_` )| |  | |/',__)/' v `\ /'_`\   |
|   | (_) )| |  ( (_| |( (_) || ( ) ( ) |( (_| || |_ | |\__, \| (˅) |( (_) )  |
|   | ,__/'(_)  `\__,_)`\__  |(_) (_) (_)`\__,_)`\__)(_)(____/(_) (_)`\___/'  |
|   | |                ( )_) |                                                |
|   (_)                 \___/'                                                |
|                                                                             |
| General Bots Copyright (c) Pragmatismo.io. All rights reserved.             |
| Licensed under the AGPL-3.0.                                                |
|                                                                             |
| According to our dual licensing model, this program can be used either      |
| under the terms of the GNU Affero General Public License, version 3,        |
| or under a proprietary license.                                             |
|                                                                             |
| The texts of the GNU Affero General Public License with an additional       |
| permission and of our proprietary license can be found at and               |
| in the LICENSE file you have received along with this program.              |
|                                                                             |
| This program is distributed in the hope that it will be useful,             |
| but WITHOUT ANY WARRANTY, without even the implied warranty of              |
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                |
| GNU Affero General Public License for more details.                         |
|                                                                             |
| "General Bots" is a registered trademark of Pragmatismo.io.                 |
| The licensing of the program under the AGPLv3 does not imply a              |
| trademark license. Therefore any rights, title and interest in              |
| our trademarks remain entirely with us.                                     |
|                                                                             |
\*****************************************************************************/
/**
 * @fileoverview General Bots server core.
 */
'use strict';
import { AuthenticationContext } from 'adal-node';
import { GBLog } from 'botlib';
import urlJoin from 'url-join';
import { AzureDeployerService } from '../../azuredeployer.gbapp/services/AzureDeployerService.js';
import { GuaribasInstance } from '../../core.gbapp/models/GBModel.js';
import { GBConfigService } from '../../core.gbapp/services/GBConfigService.js';
import { GBDeployer } from '../../core.gbapp/services/GBDeployer.js';
import { GBImporter } from '../../core.gbapp/services/GBImporterService.js';
import { GuaribasAdmin } from '../models/AdminModel.js';
import msRestAzure from 'ms-rest-azure';
import Path from 'path';
import { caseSensitive_Numbs_SpecialCharacters_PW, lowercase_PW } from 'super-strong-password-generator';
import crypto from 'crypto';
import Fs from 'fs';
import { GBServer } from '../../../src/app.js';
import { DialogKeywords } from '../../basic.gblib/services/DialogKeywords.js';
/**
 * Services for server administration.
 */
export class GBAdminService {
    static GB_PROMPT = 'GeneralBots: ';
    static masterBotInstanceId = 0;
    static StrongRegex = new RegExp('^(?=.*[a-z])(?=.*[A-Z])(?=.*[0-9])(?=.*[!@#$%^&*+_-])(?=.{8,})');
    core;
    constructor(core) {
        this.core = core;
    }
    static generateUuid() {
        return crypto.randomUUID();
    }
    static getNodeVersion() {
        const packageJson = urlJoin(process.cwd(), 'package.json');
        const pkg = JSON.parse(Fs.readFileSync(packageJson, 'utf8'));
        return pkg.engines.node.replace('=', '');
    }
    static async getADALTokenFromUsername(username, password) {
        const credentials = await GBAdminService.getADALCredentialsFromUsername(username, password);
        return credentials.tokenCache._entries[0].accessToken;
    }
    static async getADALCredentialsFromUsername(username, password) {
        return await msRestAzure.loginWithUsernamePassword(username, password);
    }
    static getMobileCode() {
        return this.getNumberIdentifier(6);
    }
    static getRndPassword() {
        let password = caseSensitive_Numbs_SpecialCharacters_PW(15);
        password = password.replace(/[\@\[\=\:\;\?\"\'\#]/gi, '*');
        return password;
    }
    static getRndReadableIdentifier() {
        return lowercase_PW(14);
    }
    static getNumberIdentifier(digits = 14) {
        if (digits <= 0) {
            throw new Error('Number of digits should be greater than 0.');
        }
        const min = 10 ** (digits - 1);
        const max = 10 ** digits - 1;
        const randomNumber = Math.floor(Math.random() * (max - min + 1)) + min;
        return randomNumber.toString();
    }
    /**
     * @see https://stackoverflow.com/a/52171480
     */
    static getHash(str, seed = 0) {
        let h1 = 0xdeadbeef ^ seed, h2 = 0x41c6ce57 ^ seed;
        for (let i = 0, ch; i < str.length; i++) {
            ch = str.charCodeAt(i);
            h1 = Math.imul(h1 ^ ch, 2654435761);
            h2 = Math.imul(h2 ^ ch, 1597334677);
        }
        h1 = Math.imul(h1 ^ (h1 >>> 16), 2246822507) ^ Math.imul(h2 ^ (h2 >>> 13), 3266489909);
        h2 = Math.imul(h2 ^ (h2 >>> 16), 2246822507) ^ Math.imul(h1 ^ (h1 >>> 13), 3266489909);
        return 4294967296 * (2097151 & h2) + (h1 >>> 0);
    }
    static async undeployPackageCommand(text, min) {
        const packageName = text.split(' ')[1];
        const importer = new GBImporter(min.core);
        const deployer = new GBDeployer(min.core, importer);
        const path = DialogKeywords.getGBAIPath(min.botId, null, packageName);
        const localFolder = Path.join('work', path);
        await deployer.undeployPackageFromLocalPath(min.instance, localFolder);
    }
    static isSharePointPath(path) {
        return path.indexOf('sharepoint.com') !== -1;
    }
    static async deployPackageCommand(min, user, text, deployer) {
        const packageName = text.split(' ')[1];
        if (!this.isSharePointPath(packageName)) {
            const additionalPath = GBConfigService.get('ADDITIONAL_DEPLOY_PATH');
            if (additionalPath === undefined) {
                throw new Error('ADDITIONAL_DEPLOY_PATH is not set and deployPackage was called.');
            }
            await deployer['deployPackage2'](min, user, urlJoin(additionalPath, packageName));
        }
        else {
            const folderName = text.split(' ')[2];
            const packageType = Path.extname(folderName).substr(1);
            const gbaiPath = DialogKeywords.getGBAIPath(min.instance.botId, packageType, null);
            const localFolder = Path.join('work', gbaiPath);
            // .gbot packages are handled using storage API, so no download
            // of local resources is required.
            const gbai = DialogKeywords.getGBAIPath(min.instance.botId);
            await deployer['downloadFolder'](min, Path.join('work', `${gbai}`), Path.basename(localFolder));
            await deployer['deployPackage2'](min, user, localFolder);
        }
    }
    static async rebuildIndexPackageCommand(min, deployer) {
        const service = await AzureDeployerService.createInstance(deployer);
        await deployer.rebuildIndex(min.instance, service.getKBSearchSchema(min.instance.searchIndex));
    }
    static async syncBotServerCommand(min, deployer) {
        const serverName = `${min.instance.botId}-server`;
        const service = await AzureDeployerService.createInstance(deployer);
        service.syncBotServerRepository(min.instance.botId, serverName);
    }
    async setValue(instanceId, key, value) {
        const options = { where: {} };
        options.where = { key: key, instanceId: instanceId };
        let admin = await GuaribasAdmin.findOne(options);
        if (admin === null) {
            admin = new GuaribasAdmin();
            admin.key = key;
        }
        admin.value = value;
        admin.instanceId = instanceId;
        await admin.save();
    }
    async updateSecurityInfo(instanceId, authenticatorTenant, authenticatorAuthorityHostUrl) {
        const options = { where: {} };
        options.where = { instanceId: instanceId };
        const item = await GuaribasInstance.findOne(options);
        item.authenticatorTenant = authenticatorTenant;
        item.authenticatorAuthorityHostUrl = authenticatorAuthorityHostUrl;
        return item.save();
    }
    async getValue(instanceId, key) {
        const options = { where: {} };
        options.where = { key: key, instanceId: instanceId };
        const obj = await GuaribasAdmin.findOne(options);
        return obj.value;
    }
    async acquireElevatedToken(instanceId, root = false) {
        if (root) {
            const minBoot = GBServer.globals.minBoot;
            instanceId = minBoot.instance.instanceId;
        }
        GBLog.info(`Acquiring token for instanceId: ${instanceId} (root: ${root}).`);
        let expiresOnV;
        try {
            expiresOnV = await this.getValue(instanceId, 'expiresOn');
        }
        catch (error) {
            throw new Error(`/setupSecurity is required before running /publish.`);
        }
        return new Promise(async (resolve, reject) => {
            const instance = await this.core.loadInstanceById(instanceId);
            const expiresOn = new Date(expiresOnV);
            if (expiresOn.getTime() > new Date().getTime()) {
                const accessToken = await this.getValue(instanceId, 'accessToken');
                resolve(accessToken);
            }
            else {
                const authorizationUrl = urlJoin(instance.authenticatorAuthorityHostUrl, instance.authenticatorTenant, '/oauth2/authorize');
                const refreshToken = await this.getValue(instanceId, 'refreshToken');
                const resource = 'https://graph.microsoft.com';
                const authenticationContext = new AuthenticationContext(authorizationUrl);
                authenticationContext.acquireTokenWithRefreshToken(refreshToken, instance.marketplaceId, instance.marketplacePassword, resource, async (err, res) => {
                    if (err !== null) {
                        reject(err);
                    }
                    else {
                        const token = res;
                        try {
                            await this.setValue(instanceId, 'accessToken', token.accessToken);
                            await this.setValue(instanceId, 'refreshToken', token.refreshToken);
                            await this.setValue(instanceId, 'expiresOn', token.expiresOn.toString());
                            resolve(token.accessToken);
                        }
                        catch (error) {
                            reject(err);
                        }
                    }
                });
            }
        });
    }
    async publish(min, packageName, republish) { }
}
//# sourceMappingURL=GBAdminService.js.map